﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Diagnostics;           // Debug.WriteLine出力用
using NDde.Server;                  // DdeServer使用
using RTedge;   // RT-edge system API
using System.Collections.Concurrent;

namespace EgDDEsvr
{
    /// <summary>
    /// EgTagの各データ型に対応するためのStruct定義
    /// </summary>
    [StructLayout(LayoutKind.Explicit)]
    public struct TagVar
    {
        [FieldOffset(0)] public bool    boVal;
        [FieldOffset(0)] public sbyte   sbVal;
        [FieldOffset(0)] public byte    byVal;
        [FieldOffset(0)] public short   shVal;
        [FieldOffset(0)] public ushort  usVal;
        [FieldOffset(0)] public int     inVal;
        [FieldOffset(0)] public uint    uiVal;
        [FieldOffset(0)] public long    loVal;
        [FieldOffset(0)] public ulong   ulVal;
        [FieldOffset(0)] public float   flVal;
        [FieldOffset(0)] public double  doVal;
    }

    //
    // DDEServerクラス
    //
    public class EgDDEServer : DdeServer                            // DdeServerを継承
    {
        private System.Timers.Timer _AdviseLoopTimer;               // Adviseタグ監視用タイマ
        private edgeCode EgProc;                                    // edgeAPI用
        public IDictionary<IntPtr, DdeInfo> DdeCliInfo;             // 接続情報リスト
        public IDictionary<string, AdviseTagInfo> AdviseTags;       // 監視タグリスト

#if DEBUG       // TODO: [DEBUG]処理時間計測
        public Stopwatch m_SW_Adv = new Stopwatch();
        public Stopwatch m_SW_Pok = new Stopwatch();
        public Stopwatch m_SW_Req = new Stopwatch();
        uint AdvCount = 0;
        uint PokCount = 0;
        uint ReqCount = 0;
        uint measureNum = 1;      // タグ点数
        List<double> AdvTimeList = new List<double>();
        List<double> PokTimeList = new List<double>();
        List<double> ReqTimeList = new List<double>();
#endif

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="service"></param>
        public EgDDEServer(string service, edgeCode egProc)
            : base(service)                                 // サービス名を指定
        {
            DdeCliInfo = new ConcurrentDictionary<IntPtr, DdeInfo>();       // 接続情報リスト初期化
            AdviseTags = new ConcurrentDictionary<string, AdviseTagInfo>(); // 監視タグリスト初期化
            EgProc = egProc;
        }

        #region NDde.lib override module

        /// <summary>
        /// サーバー起動時呼ばれる関数
        /// </summary>
        public override void Register()
        {
            base.Register();
        }
        /// <summary>
        /// サーバー終了時呼ばれる関数
        /// </summary>
        public override void Unregister()
        {
            base.Unregister();
        }
        /// <summary>
        /// クライアントからコネクトがあった場合最初に呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="topic"></param>
        /// <returns></returns>
        protected override bool OnBeforeConnect(string topic)
        {
            if (!CheckTopicName(topic)) { return false; }   // トピック名不一致の場合はエラー返却

            return true;
        }
        /// <summary>
        /// クライアントからコネクトがあったあと終わりに呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="conversation"></param>
        protected override void OnAfterConnect(DdeConversation conversation)
        {
            try
            {
                // 接続情報リストへクライアント情報を追加
                AddCliInfo(conversation);

                // サービスプロパティタグ(DDEクライアントからの接続開始回数)を更新
                EgProc.MyTags.DdeCliOpen++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_ConnectedOpen, EgProc.MyTags.DdeCliOpen);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnAfterConnect() error!! AppName:{0} Topic:{1} Handle:{2}", conversation.Service, conversation.Topic, conversation.Handle));
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnAfterConnect() error message:{0}", ex.Message));
            }
        }
        /// <summary>
        /// クライアントからDisconnectされた場合に呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="conversation"></param>
        protected override void OnDisconnect(DdeConversation conversation)
        {
            try
            {
                // クライアントの情報削除
                ClearCliInfo(conversation.Handle);

                // サービスプロパティタグ(DDEクライアントからの接続終了回数)を更新
                EgProc.MyTags.DdeCliClose++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_ConnectedClose, EgProc.MyTags.DdeCliClose);

#if DEBUG       // TODO: [DEBUG]処理時間計測
                measureProc();
#endif
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnDisconnect() error!! AppName:{0} Topic:{1} Handle:{2}", conversation.Service, conversation.Topic, conversation.Handle));
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnDisconnect() error message:{0}", ex.Message));
            }
        }
        /// <summary>
        /// Adviseがスタートした時に呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="conversation"></param>
        /// <param name="item"></param>
        /// <param name="format"></param>
        /// <returns></returns>
        protected override bool OnStartAdvise(DdeConversation conversation, string item, int format)
        {
            try
            {
                // パラメータチェック
                if (!CheckFormat(format)) { return false; }     // フォーマット不一致の場合はエラー返却
                if (!CheckTopicName(conversation.Topic)) { return false; }  // トピック名不一致の場合はエラー返却
                if (!CheckItemName(item)) { return false; }     // アイテム名が未指定の場合はエラー返却

                // 接続情報リスト検索
                if (DdeCliInfo.ContainsKey(conversation.Handle) == false)
                {
                    // 接続情報がないクライアントからの要求はエラーとする
                    EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_CLIENT_NOINFO);   // サービスプロパティタグ(最新のエラーコード)を更新
                    return false;
                }

                // 監視タグリストへ登録
                bool ret = AddAdviseTagInfo(conversation, item);

                if (ret == true)
                {
                    // クライアント別の監視タグを管理するため、接続情報リストへタグ登録
                    DdeCliInfo[conversation.Handle].AddItem(item);
                    // Adviseタグ監視用タイマ初期化
                    InitAdviseLoopTimer();
                }

                return ret;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnStartAdvise() error!! item:{0} format:{1}", item, format));
                return false;
            }
        }
        /// <summary>
        /// AdviseがStopしたときに呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="conversation"></param>
        /// <param name="item"></param>
        protected override void OnStopAdvise(DdeConversation conversation, string item)
        {
            try
            {
                // パラメータチェック
                if (!CheckTopicName(conversation.Topic)) { return; }            // トピック名不一致の場合は無処理
                if (!DdeCliInfo.ContainsKey(conversation.Handle)) { return; }   // クライアント情報がない場合は無処理
                if (!CheckItemName(item)) { return; }                           // アイテム名が未指定の場合は無処理

                // 接続情報リストの指定タグ情報を削除
                DdeCliInfo[conversation.Handle].DelItem(item);

                // 監視タグリストの指定タグ情報を削除
                ClearAdviseTagInfo(conversation.Handle, item);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnStopAdvise() error!! item:{0}", item));
            }
        }
        /// <summary>
        /// コマンド実行メッセージが来た時に呼ばれる関数
        /// 必要に応じて処理追加
        /// </summary>
        /// <param name="conversation"></param>
        /// <param name="command"></param>
        /// <returns></returns>
        protected override ExecuteResult OnExecute(DdeConversation conversation, string command)
        {
            // NOTE: コマンド非対応

            return ExecuteResult.Processed;
        }
        /// <summary>
        /// Pokeメッセージが来た時に呼ばれる関数
        /// </summary>
        /// <param name="conversation"></param>
        /// <param name="item"></param>
        /// <param name="data"></param>
        /// <param name="format"></param>
        /// <returns></returns>
        protected override PokeResult OnPoke(DdeConversation conversation, string item, byte[] data, int format)
        {
            try
            {
#if DEBUG       // TODO: [DEBUG]処理時間計測
                if (PokCount == 0) { m_SW_Pok.Restart(); }
#endif

                // パラメータチェック
                if (!CheckFormat(format)) { return PokeResult.NotProcessed; }                   // フォーマット不一致の場合はエラー返却
                if (!CheckTopicName(conversation.Topic)) { return PokeResult.NotProcessed; }    // トピック名不一致の場合はエラー返却
                if (!CheckItemName(item)) { return PokeResult.NotProcessed; }                   // アイテム名が未指定の場合はエラー返却

                // サービスプロパティタグ(DDEクライアントからの書き込み要求回数)を更新
                EgProc.MyTags.DdeWriteReqCount++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_WriteReqCount, EgProc.MyTags.DdeWriteReqCount);

                int result;
                object type = new object();
                // データタイプ取得
                result = EgProc.EGAPI.EgTagGetProperty(item, "Type", ref type);
                if (result != 0)
                {
                    // プロパティが取得できない場合はエラー返却
                    EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_TAG_NOEXIST); // サービスプロパティタグ(最新のエラーコード)を更新
                    return PokeResult.NotProcessed;
                }

                string text = System.Text.Encoding.ASCII.GetString(data);      // バイト配列を文字列に変換

                // 取得した文字列から各データ型に変換してEgTagに書き込み
                TagVar varVal = new TagVar();
                switch ((ushort)type)
                {
                    case (ushort)EGDEFINE.egTagDataType.Boolean:    if(bool.TryParse(text, out varVal.boVal))   { EgProc.EGAPI.EgTagWrite(item, varVal.boVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.SByte:      if(sbyte.TryParse(text, out varVal.sbVal))  { EgProc.EGAPI.EgTagWrite(item, varVal.sbVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Byte:       if(byte.TryParse(text, out varVal.byVal))   { EgProc.EGAPI.EgTagWrite(item, varVal.byVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Int16:      if(short.TryParse(text, out varVal.shVal))  { EgProc.EGAPI.EgTagWrite(item, varVal.shVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.UInt16:     if(ushort.TryParse(text, out varVal.usVal)) { EgProc.EGAPI.EgTagWrite(item, varVal.usVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Int32:      if(int.TryParse(text, out varVal.inVal))    { EgProc.EGAPI.EgTagWrite(item, varVal.inVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.UInt32:     if(uint.TryParse(text, out varVal.uiVal))   { EgProc.EGAPI.EgTagWrite(item, varVal.uiVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Int64:      if(long.TryParse(text, out varVal.loVal))   { EgProc.EGAPI.EgTagWrite(item, varVal.loVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.UInt64:     if(ulong.TryParse(text, out varVal.ulVal))  { EgProc.EGAPI.EgTagWrite(item, varVal.ulVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Float:      if(float.TryParse(text, out varVal.flVal))  { EgProc.EGAPI.EgTagWrite(item, varVal.flVal); } break;
                    case (ushort)EGDEFINE.egTagDataType.Double:     if(double.TryParse(text, out varVal.doVal)) { EgProc.EGAPI.EgTagWrite(item, varVal.doVal); } break;
                    default: break;
                }

#if DEBUG       // TODO: [DEBUG]処理時間計測
                PokCount++;
                if (PokCount == measureNum) { PokTimeList.Add(m_SW_Pok.Elapsed.TotalMilliseconds); PokCount = 0; }
#endif
                return PokeResult.Processed;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnPoke() error!! item:{0}", item));
                return PokeResult.NotProcessed;
            }
        }
        /// <summary>
        /// Requestメッセージが来た時に呼ばれる関数
        /// クライアントにデータを返す
        /// </summary>
        /// <param name="conversation"></param>
        /// <param name="item"></param>
        /// <param name="format"></param>
        /// <returns></returns>
        protected override RequestResult OnRequest(DdeConversation conversation, string item, int format)
        {
            try
            {
#if DEBUG       // TODO: [DEBUG]処理時間計測
                if (ReqCount == 0) { m_SW_Req.Restart(); }
#endif
                // パラメータチェック
                if (!CheckFormat(format)) { return RequestResult.NotProcessed; }                // フォーマット不一致の場合はエラー返却
                if (!CheckTopicName(conversation.Topic)) { return RequestResult.NotProcessed; } // トピック名不一致の場合はエラー返却
                if (!CheckItemName(item)) { return RequestResult.NotProcessed; }                // アイテム名が未指定の場合はエラー返却

                // サービスプロパティタグ(DDEクライアントからの読み込み要求回数)を更新
                EgProc.MyTags.DdeReadReqCount++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_ReadReqCount, EgProc.MyTags.DdeReadReqCount);

                string ret = "";
                Int32 result;

                // データタイプ取得
                object type = new object();
                object value = new object();
                result = EgProc.EGAPI.EgTagGetProperty(item, "Type", ref type);
                if (result == 0)
                {
                    // EgTagRead
                    EgProc.EGAPI.EgTagRead(item, ref value);

                    // 各タグのデータ型に基づき、値を文字列へ変換
                    switch ((UInt16)type)
                    {
                        case (ushort)EGDEFINE.egTagDataType.Boolean:    var val0 = (Boolean)value;  ret = val0.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.SByte:      var val1 = (SByte)value;    ret = val1.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Byte:       var val2 = (Byte)value;     ret = val2.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Int16:      var val3 = (Int16)value;    ret = val3.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.UInt16:     var val4 = (UInt16)value;   ret = val4.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Int32:      var val5 = (Int32)value;    ret = val5.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.UInt32:     var val6 = (UInt32)value;   ret = val6.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Int64:      var val7 = (Int64)value;    ret = val7.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.UInt64:     var val8 = (UInt64)value;   ret = val8.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Float:      var val9 = (float)value;    ret = val9.ToString(); break;
                        case (ushort)EGDEFINE.egTagDataType.Double:     var val10 = (double)value;  ret = val10.ToString(); break;
                        default: break;
                    }
                }

                // バイト配列に変換して返却
                RequestResult rr = new RequestResult(System.Text.Encoding.Default.GetBytes(ret + "\0"));

#if DEBUG       // TODO: [DEBUG]処理時間計測
                ReqCount++;
                if (ReqCount == measureNum) { ReqTimeList.Add(m_SW_Req.Elapsed.TotalMilliseconds); ReqCount = 0; }
#endif
                return rr;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnRequest() error!! item:{0} format:{1}", item, format));
                return RequestResult.NotProcessed;
            }
        }
        /// <summary>
        /// Advise関数
        /// クライアントにデータを渡す
        /// </summary>
        /// <param name="topic"></param>
        /// <param name="item"></param>
        /// <param name="format"></param>
        /// <returns></returns>
        protected override byte[] OnAdvise(string topic, string item, int format)
        {
            byte[] output = null;
            try
            {
                // パラメータチェック
                if (!CheckFormat(format)) { return output; }            // フォーマット不一致の場合はNull返却
                if (!CheckTopicName(topic)) { return output; }          // トピック名不一致の場合はNull返却
                if (!CheckItemName(item)) { return output; }            // アイテム名が未指定の場合はNull返却
                if (!AdviseTags.ContainsKey(item)) { return output; }   // 監視タグリストに存在しない場合はNull返却

                // 監視タグリストのカレント値を返却する
                output = System.Text.Encoding.ASCII.GetBytes(AdviseTags[item].StrData + "\0");

#if DEBUG       // TODO: [DEBUG]処理時間計測
                AdvCount++;
                if (AdvCount == measureNum) { AdvTimeList.Add(m_SW_Adv.Elapsed.TotalMilliseconds); AdvCount = 0; }
#endif

                return output;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, String.Format("OnAdvise() error!! item:{0} format:{1}", item, format));
                return output;
            }
        }

        #endregion

        #region public

        /// <summary>
        /// DDEサーバー開始処理
        /// </summary>
        /// <returns></returns>
        public bool Start()
        {
            try
            {
                Register();                 // DDE Server 開始
                return IsRegistered;        // 処理結果返却
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                return false;
            }
        }

        /// <summary>
        /// DDEサーバー終了処理
        /// </summary>
        public void End()
        {
            try
            {
                Unregister();           // DDE Server 終了

                IntervalState(false);   // タグ監視用タイマー停止
                DdeCliInfo.Clear();     // 接続情報リスト削除
                AdviseTags.Clear();     // 監視タグリスト削除
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        #endregion

        #region private
        /// <summary>
        /// タグ監視用タイマー初期化
        /// </summary>
        private void InitAdviseLoopTimer()
        {
            // タイマー未初期化、かつ監視タグリストに登録がある場合のみ処理
            if (_AdviseLoopTimer == null && AdviseTags.Count > 0)
            {
                _AdviseLoopTimer = new System.Timers.Timer();
                _AdviseLoopTimer.Elapsed += this.OnTimerElapsed;
                _AdviseLoopTimer.Interval = EgProc.MyTags.Cycle;
                _AdviseLoopTimer.Enabled = true;
            }
        }

        /// <summary>
        /// タグ監視用タイマーのスイッチ
        /// </summary>
        /// <param name="state"></param>
        private void IntervalState(bool state)
        {
            if (_AdviseLoopTimer != null)
            {
                _AdviseLoopTimer.Enabled = state;
            }
        }

        /// <summary>
        /// タグ監視処理
        /// NOTE: 指定周期でタグ値を監視し、データ更新があった場合にAdviseを呼ぶ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private void OnTimerElapsed(object sender, System.Timers.ElapsedEventArgs args)
        {
            try
            {
                if (AdviseTags.Count == 0) { return; }  // タグ監視リストに登録がない場合は無処理で抜ける

#if DEBUG       // TODO: [DEBUG]処理時間計測
                AdvCount = 0;
                m_SW_Adv.Restart();
#endif
                _AdviseLoopTimer.Enabled = false;
                var sw = new Stopwatch();
                sw.Start();

                // タグ監視処理の稼働カウンタ更新
                EgProc.MyTags.ReadLive++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_READ_Live, EgProc.MyTags.ReadLive);

                object value = new object();
                // 監視タグリスト内のEgTagを読み込み
                foreach (var item in AdviseTags)
                {
                    EgProc.EGAPI.EgTagRead(item.Key, ref value);

                    // タグ値更新
                    if (item.Value.UpdateData(value) == true)
                    {
                        // タグ値が変化した場合、Advise()コール
                        Advise(EgProc.MyTags.TopicName, item.Key);

                        // サービスタグ(Adviseコール回数)を更新
                        EgProc.MyTags.AdviseCount++;
                    }
                }
                // サービスタグ(Adviseコール回数)を書き込み
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_READ_AdviseCount, EgProc.MyTags.AdviseCount);

                sw.Stop();

                // 処理時間更新
                EgProc.MyTags.Update_OnAdviseTime((float)sw.Elapsed.TotalMilliseconds);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_ERR, "OnTimerElapsed() Error!!  " + ex.Message);
            }
            _AdviseLoopTimer.Enabled = true;
        }

        /// <summary>
        /// 接続情報リストへクライアント情報を追加
        /// </summary>
        /// <param name="conversation"></param>
        private void AddCliInfo(DdeConversation conversation)
        {
            // 接続情報リスト検索
            // NOTE: クライアントの識別はハンドル値で行う
            if (DdeCliInfo.ContainsKey(conversation.Handle) == false)
            {
                // (未登録の場合)接続情報リストへ追加
                DdeCliInfo.Add(conversation.Handle, new DdeInfo() { Handle = conversation.Handle, Service = conversation.Service, Topic = conversation.Topic, RefCount = 1 });

                // サービスプロパティタグ(接続クライアント数)を更新
                EgProc.MyTags.DdeClientNum++;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_CliNum, EgProc.MyTags.DdeClientNum);
            }
            else
            {
                // (登録済みの場合)参照数をインクリメント
                DdeCliInfo[conversation.Handle].RefCount++;
            }
        }

        /// <summary>
        /// 指定クライアントの情報削除
        /// </summary>
        /// <param name="pCli">クライアントのハンドル値</param>
        private void ClearCliInfo(IntPtr pCli)
        {
            // 接続情報リストにクライアント未登録の場合は、無処理で抜ける
            if (DdeCliInfo.ContainsKey(pCli) == false) { return; }

            // ----------------------------------------
            // 監視タグリストからクライアントの情報削除
            // ----------------------------------------
            // クライアントがAdvise利用中か判定
            if (DdeCliInfo[pCli].AdviseItem != null && DdeCliInfo[pCli].AdviseItem.Count > 0)
            {
                // Advise利用中の場合、監視タグリストから情報削除
                // NOTE: 本ケースに入るのは、クライアントからStopAdviseがコールされずに切断された場合です。
                foreach (var item in DdeCliInfo[pCli].AdviseItem)
                {
                    if (AdviseTags.ContainsKey(item))
                    {
                        if (AdviseTags[item].Client.Count == 1)
                        {
                            // 該当タグの監視を他のクライアントが行っていない場合は、リストから削除
                            AdviseTags.Remove(item);
                        }
                        else
                        {
                            // 該当タグの監視を他のクライアントが行っている場合は、クライアント情報のみ削除
                            AdviseTags[item].DelClient(pCli);
                        }
                    }
                }
                // サービスプロパティタグ(現在の監視アイテム数)を更新
                EgProc.MyTags.DdeMoniterNum = (uint)AdviseTags.Count;
                EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_MoniterNum, EgProc.MyTags.DdeMoniterNum);
            }

            // ----------------------------------------
            // 接続情報リストからクライアントの情報削除
            // ----------------------------------------
            if (DdeCliInfo[pCli].RefCount > 1)
            {
                // 同一ハンドルで複数接続状態の場合、参照数をデクリメント
                DdeCliInfo[pCli].RefCount--;
            }
            else
            {
                // 接続情報リストから削除
                DdeCliInfo.Remove(pCli);
            }
            // サービスプロパティタグ(接続クライアント数)を更新
            EgProc.MyTags.DdeClientNum = (uint)DdeCliInfo.Count;
            EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_CliNum, EgProc.MyTags.DdeClientNum);
        }


        /// <summary>
        /// 監視タグリストへ登録
        /// </summary>
        /// <param name="conversation">クライアント情報</param>
        /// <param name="item">タグ名</param>
        /// <returns></returns>
        private bool AddAdviseTagInfo(DdeConversation conversation, string item)
        {
            bool ret = true;

            // 監視タグリスト検索
            if (AdviseTags.ContainsKey(item) == false)
            {
                // 未登録の場合
                // タグ情報取得
                Int32 result;
                object type = new object();
                result = EgProc.EGAPI.EgTagGetProperty(item, "Type", ref type);     // EgTagのデータ型取得
                if (result == 0)                                                    // 取得正常の場合
                {
                    object value = new object();
                    EgProc.EGAPI.EgTagRead(item, ref value);                        // EgTagRead
                    AdviseTags.Add(item, new AdviseTagInfo(item, (UInt16)type, value, conversation.Handle));   // 監視タグリストへ追加
                    Advise(conversation.Topic, item);                               // 初回は必ずAdvise()をコールする

                    EgProc.MyTags.DdeMoniterReqCount++;                             // サービスプロパティタグ(DDEクライアントからの監視要求回数)を更新
                    EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_MoniterReqCount, EgProc.MyTags.DdeMoniterReqCount);
                    EgProc.MyTags.DdeMoniterNum = (uint)AdviseTags.Count;           // サービスプロパティタグ(現在の監視アイテム数)を更新
                    EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_MoniterNum, EgProc.MyTags.DdeMoniterNum);
                }
                else
                {
                    EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_TAG_NOEXIST); // サービスプロパティタグ(最新のエラーコード)を更新
                    ret = false;    // エラー返却
                }
            }
            else
            {
                // すでに登録済みの場合は、クライアント識別情報のみ追加登録する
                AdviseTags[item].AddClient(conversation.Handle);
            }
            return ret;
        }

        /// <summary>
        /// 監視タグリストから指定タグの情報を削除
        /// </summary>
        /// <param name="pCli">クライアントのハンドル値</param>
        private void ClearAdviseTagInfo(IntPtr pCli, string item)
        {
            if (AdviseTags.ContainsKey(item) == true)
            {
                if (AdviseTags[item].Client.Count == 1)
                {
                    // 監視タグリストから指定タグ情報を削除
                    AdviseTags.Remove(item);

                    // サービスプロパティタグ(現在の監視アイテム数)を更新
                    EgProc.MyTags.DdeMoniterNum = (uint)AdviseTags.Count;
                    EgProc.EGAPI.EgTagWrite(EgProc.EGCONST.TAG_MYSERV_DDE_MoniterNum, EgProc.MyTags.DdeMoniterNum);
                }
                else
                {
                    // 監視タグリストから指定タグ情報内のクライアントハンドル値のみを削除
                    // NOTE: 本ケースは、他のクライアントがこのタグの監視を継続している場合を想定
                    AdviseTags[item].DelClient(pCli);
                }
            }
        }

        /// <summary>
        /// フォーマット判定処理
        /// </summary>
        /// <param name="format"></param>
        /// <returns>true:判定OK, false:判定NG</returns>
        private bool CheckFormat(int format)
        {
            if (format != 1)        // CF_TEXT(1)のみ対応
            {
                EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_FORMAT); // サービスプロパティタグ(最新のエラーコード)を更新
                return false;
            }
            return true;
        }

        /// <summary>
        /// トピック名判定処理
        /// </summary>
        /// <param name="topic"></param>
        /// <returns>true:判定OK, false:判定NG</returns>
        private bool CheckTopicName(string topic)
        {
            // トピック名が設定値と異なる場合は、エラーコードを登録しエラー返却します
            if (topic != EgProc.MyTags.TopicName)
            {
                EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_TOPICNAME);   // サービスプロパティタグ(最新のエラーコード)を更新
                return false;
            }
            return true;
        }

        /// <summary>
        /// アイテム名判定処理
        /// </summary>
        /// <param name="item"></param>
        /// <returns>true:判定OK, false:判定NG</returns>
        private bool CheckItemName(string item)
        {
            // ここではブランク判定のみ
            if (string.IsNullOrEmpty(item))
            {
                EgProc.MyTags.SetLastError(EDGECONST.ERR_EGDDESVR_ITEMNAME);    // サービスプロパティタグ(最新のエラーコード)を更新
                return false;
            }
            return true;
        }


#if DEBUG       // TODO: [DEBUG]処理時間計測
        /// <summary>
        /// [DEBUG]処理時間集計
        /// </summary>
        private void measureProc()
        {
            double sum, min, max, avg, work;
            // Poke処理
            if (PokTimeList.Count > 0)
            {
                sum = 0; min = 0; max = 0; avg = 0; work = 0;
                for (int i = 0; i < PokTimeList.Count; i++)
                {
                    work = PokTimeList[i];
                    sum += work;
                    if (min == 0) { min = work; }
                    if (min > work) { min = work; }
                    if (max < work) { max = work; }
                }
                avg = sum / PokTimeList.Count;
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_OPE, String.Format("OnPoke   time avg:{0} min:{1} max:{2}", avg.ToString("F4"), min.ToString("F4"), max.ToString("F4")));
                PokTimeList.Clear();
                m_SW_Pok.Reset();
                PokCount = 0;
            }
            // Request処理
            if (ReqTimeList.Count > 0)
            {
                sum = 0; min = 0; max = 0; avg = 0; work = 0;
                for (int i = 0; i < ReqTimeList.Count; i++)
                {
                    work = ReqTimeList[i];
                    sum += work;
                    if (min == 0) { min = work; }
                    if (min > work) { min = work; }
                    if (max < work) { max = work; }
                }
                avg = sum / ReqTimeList.Count;
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_OPE, String.Format("OnReqest time avg:{0} min:{1} max:{2}", avg.ToString("F4"), min.ToString("F4"), max.ToString("F4")));
                ReqTimeList.Clear();
                m_SW_Req.Reset();
                ReqCount = 0;
            }
            // Advise処理
            if (AdvTimeList.Count > 0)
            {
                sum = 0; min = 0; max = 0; avg = 0; work = 0;
                for (int i = 0; i < AdvTimeList.Count; i++)
                {
                    work = AdvTimeList[i];
                    sum += work;
                    if (min == 0) { min = work; }
                    if (min > work) { min = work; }
                    if (max < work) { max = work; }
                }
                avg = sum / AdvTimeList.Count;
                edge_FW._PrintSysLog(EgProc.EGCONST.SERVICENAME, edge_FW.EG_LOG_OPE, String.Format("OnAdvise time avg:{0} min:{1} max:{2}", avg.ToString("F4"), min.ToString("F4"), max.ToString("F4")));
                AdvTimeList.Clear();
                m_SW_Adv.Reset();
                AdvCount = 0;
            }
        }
#endif

#endregion
    }
}
